@@ -24,7 +24,7 @@
 #include "mpi-internal.h"
 
 int
-gcry_mpi_cmp_ui (gcry_mpi_t u, unsigned long v)
+_gcry_mpi_cmp_ui (gcry_mpi_t u, unsigned long v)
 {
   mpi_limb_t limb = v;
 
@@ -53,38 +53,75 @@ gcry_mpi_cmp_ui (gcry_mpi_t u, unsigned long v)
     return 1;
 }
 
-int
-gcry_mpi_cmp (gcry_mpi_t u, gcry_mpi_t v)
+
+/* Helper for _gcry_mpi_cmp and _gcry_mpi_cmpabs.  */
+static int
+do_mpi_cmp (gcry_mpi_t u, gcry_mpi_t v, int absmode)
 {
   mpi_size_t usize;
   mpi_size_t vsize;
+  int usign;
+  int vsign;
   int cmp;
 
-  _gcry_mpi_normalize (u);
-  _gcry_mpi_normalize (v);
+  if (mpi_is_opaque (u) || mpi_is_opaque (v))
+    {
+      /* We have no signan and thus ABSMODE has no efeect here.  */
+      if (mpi_is_opaque (u) && !mpi_is_opaque (v))
+        return -1;
+      if (!mpi_is_opaque (u) && mpi_is_opaque (v))
+        return 1;
+      if (!u->sign && !v->sign)
+        return 0; /* Empty buffers are identical.  */
+      if (u->sign < v->sign)
+        return -1;
+      if (u->sign > v->sign)
+        return 1;
+      return memcmp (u->d, v->d, (u->sign+7)/8);
+    }
+  else
+    {
+      _gcry_mpi_normalize (u);
+      _gcry_mpi_normalize (v);
 
-  usize = u->nlimbs;
-  vsize = v->nlimbs;
+      usize = u->nlimbs;
+      vsize = v->nlimbs;
+      usign = absmode? 0 : u->sign;
+      vsign = absmode? 0 : v->sign;
 
-  /* Compare sign bits.  */
+      /* Compare sign bits.  */
 
-  if (!u->sign && v->sign)
-    return 1;
-  if (u->sign && !v->sign)
-    return -1;
+      if (!usign && vsign)
+        return 1;
+      if (usign && !vsign)
+        return -1;
 
-  /* U and V are either both positive or both negative.  */
-
-  if( usize != vsize && !u->sign && !v->sign )
-    return usize - vsize;
-  if( usize != vsize && u->sign && v->sign )
-    return vsize + usize;
-  if( !usize )
-    return 0;
-  if( !(cmp = _gcry_mpih_cmp( u->d, v->d, usize )) )
-    return 0;
-  if( (cmp < 0?1:0) == (u->sign?1:0))
-    return 1;
+      /* U and V are either both positive or both negative.  */
 
+      if (usize != vsize && !usign && !vsign)
+        return usize - vsize;
+      if (usize != vsize && usign && vsign)
+        return vsize + usize;
+      if (!usize )
+        return 0;
+      if (!(cmp = _gcry_mpih_cmp (u->d, v->d, usize)))
+        return 0;
+      if ((cmp < 0?1:0) == (usign?1:0))
+        return 1;
+    }
   return -1;
 }
+
+
+int
+_gcry_mpi_cmp (gcry_mpi_t u, gcry_mpi_t v)
+{
+  return do_mpi_cmp (u, v, 0);
+}
+
+/* Compare only the absolute values.  */
+int
+_gcry_mpi_cmpabs (gcry_mpi_t u, gcry_mpi_t v)
+{
+  return do_mpi_cmp (u, v, 1);
+}
